/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 General AVL balanced binary tree routines: insertion,
 remotion, search, tree traversal.

 $Id: avl.c,v 1.1 2001/01/17 11:20:10 loreti Exp $

 *------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "avl.h"

/**
 | Local procedure prototypes
**/

static node *balanceNode(node *);
static node *insertNode(node *, item *, node **,
                        int(const item *, const item *), int *);
static node *newNode(item *);
static node *removeLeftmost(node *, node **);
static node *removeNode(node *, const item *,
                        int(const item *, const item *), void(item *),
                        int *);
static node *restoreLeftBalance(node *, const int);
static node *restoreRightBalance(node *, const int);
static node *singleRotateLeft(node *);
static node *singleRotateRight(node *);

item *findAvl(
  node       *root,
  const item *pTarget,
  int         compare(const item *, const item *)
) {

/**
 | "findAvl" explores the subtree rooted at *root, searching for a
 | node comparing equal to the one in *pTarget.  Returns a pointer to
 | the node, or zero if it has not been found.
**/

  if (root) {
    int c = compare(pTarget, root->pItem);

    if (c < 0) {
      return findAvl(root->left, pTarget, compare);
    } else if (c > 0) {
      return findAvl(root->right, pTarget, compare);
    } else {
      return root->pItem;
    }
  } else {
    return 0;
  }
}

int insertAvl(
  node **pRoot,
  item  *pItem,
  node **ppNode,
  int    compare(const item *, const item *)
) {

/**
 | "insertAvl" tries to insert in the binary tree rooted at **pRoot a
 | node containing the user data in *pItem.  The return value is OK in
 | case of success; if a node with the given item already exists, the
 | tree is unchanged and the return value is ERROR.  In both cases, in
 | *pNode will be stored a pointer to the node containing the given
 | item.
 |
 | "compare" is a function taking two pointers to data items, that
 | returns a negative value if the first one is lesser that the second
 | one; a positive value if the first one is bigger than the second
 | one; or zero if the two items have the same value.
**/

  int retVal = OK;

/**
 | Consider a special case: the tree is empty; if not, iterate down
 | the children subtrees.
**/

  if (*pRoot == 0) {
    *pRoot = *ppNode = newNode(pItem);
  } else {
    *pRoot = insertNode(*pRoot, pItem, ppNode, compare, &retVal);
  }

  return retVal;
}

void removeAvl(
  node *root,
  void delete(item *)
) {

/**
 | "removeAvl" removes the subtree rooted at *root, calling for every
 | user data item the procedure "delete" (if supplied)
**/

  if (root) {
    removeAvl(root->left, delete);
    removeAvl(root->right, delete);
    if (delete) delete(root->pItem);
    free(root);
  }
}

int removeFromAvl(
  node       **pRoot,
  const item  *pTarget,
  int          compare(const item *, const item *),
  void         delete(item *)
) {

  int retVal = ERROR;

/**
 | "removeFromAvl" explores the balanced binary tree rooted at
 | **pRoot, searching for a node containing an user data item
 | comparing equal to *ptarget; the node is then removed from the
 | tree, after having called the procedure "delete" (if supplied).
 |
 | The returned value is OK if successful, or ERROR if the wanted
 | node does not exist.
**/

  if (*pRoot != 0) {
    *pRoot = removeNode(*pRoot, pTarget, compare, delete, &retVal);
  }
  return retVal;
}

void traverseAvl(
  const node  *root,
  const order  o,
  void         fun(const item *)
) {

/**
 | "traverseAvl" traverses the subtree rooted at *root, calling for
 | every user data item the supplied procedure "fun".
**/

  if (root) {
    if (o == PREORDER) fun(root->pItem);
    traverseAvl(root->left, o, fun);
    if (o == INORDER) fun(root->pItem);
    traverseAvl(root->right, o, fun);
    if (o == POSTORDER) fun(root->pItem);
  }
}

/* -------------------- Local (static) procedures -------------------- */

node *balanceNode(
  node *root
) {

/**
 | Rebalances a subtree that has gone out of balance, with a single or
 | a double rotation to the left or to the right
**/

  if (root->balance < 0) {
    if (root->left->balance <= 0) return singleRotateRight(root);
    else {
      root->left = singleRotateLeft(root->left);
      return singleRotateRight(root);
    }

  } else {
    if (root->right->balance >= 0) return singleRotateLeft(root);
    else {
      root->right = singleRotateRight(root->right);
      return singleRotateLeft(root);
    }
  }
}

static node *insertNode(
  node  *root,
  item  *pItem,
  node **ppNode,
  int    compare(const item *, const item *),
  int   *pStatus
) {
  int c = compare(pItem, root->pItem);
  int oldBalance;

/**
 | Looks in the left or the right subtree.  If the subtree is empty,
 | inserts a new node here; if not, slides down the subtree.
**/

  if (c < 0) {
    if (root->left == 0) {
      *ppNode  = root->left = newNode(pItem);
      *pStatus = OK;
      --(root->balance);
    } else {
      oldBalance = root->left->balance;
      root->left = insertNode(root->left, pItem, ppNode, compare, pStatus);
      if (root->left->balance != oldBalance   &&   root->left->balance) {
        --(root->balance);
      }
    }

  } else if (c > 0) {
    if (root->right == 0) {
      *ppNode  = root->right = newNode(pItem);
      *pStatus = OK;
      ++(root->balance);
    } else {
      oldBalance  = root->right->balance;
      root->right = insertNode(root->right, pItem, ppNode, compare, pStatus);
      if (root->right->balance != oldBalance   &&   root->right->balance) {
        ++(root->balance);
      }
    }

  } else {

    *ppNode  = root;
    *pStatus = ERROR;
  }

/**
 | Restore the subtree balance, if the root has gone unbalanced
**/

  if (root->balance < -1   ||   root->balance > 1) return balanceNode(root);
  else                                             return root;
}

static node *newNode(
  item *pItem
) {
  node *pNode;

/**
 | Creates and initializes a new node from the heap
**/

  if ((pNode = malloc(sizeof(node))) == 0) {
    fputs("insertAvl: couldn't get heap memory\n", stderr);
    exit(EXIT_FAILURE);
  }

  pNode->left    = 0;
  pNode->right   = 0;
  pNode->balance = 0;
  pNode->pItem   = pItem;

  return pNode;
}

static node *removeLeftmost(
  node  *root,
  node **ppNode
) {

  int   oldBalance;
  node *leftChild;

/**
 | Finds and removes the leftmost descendant of a given node.
 | - If this node does not have a left child, we got what we wanted;
 | - if not, we save the balance factor of the subtree; iterate; and
 |   rebalance the subtree if needed.
**/

  if ((leftChild = root->left) == 0) {
    *ppNode = root;
    return root->right;
  }

  oldBalance = leftChild->balance;
  root->left = removeLeftmost(leftChild, ppNode);
  return restoreLeftBalance(root, oldBalance);
}

static node *removeNode(
  node       *root,
  const item *pTarget,
  int         compare(const item *, const item *),
  void        delete(item *),
  int        *pStatus
) {

/**
 | "removeNode" removes a node (if any) comparing equal to *pTarget
 | from the subtree rooted at *pNode; returns the address of the new
 | root, and stores in *pStatus OK if successful or ERROR if the
 | wanted node does not exist.
**/

  int oldBalance;
  int c = compare(pTarget, root->pItem);

  if (c == 0) {
    node *newRoot;

/**
 | We got the right node!
 |
 | - special case: if less than two children, return the only one (or
 |   a NULL link);
 | - general case: disconnect from the subtree the leftmost child of
 |   the right child, returning its address in "newRoot"; reconnect
 |   newRoot in place of the current node; rebalances the tree if
 |   needed.
 | Then we may effectively remove the node; "newRoot" contains the
 | address to be returned.
**/


    if (root->left == 0)  {
      newRoot = root->right;
    } else if (root->right == 0) {
      newRoot = root->left;
    } else {
      oldBalance  = root->right->balance;
      root->right = removeLeftmost(root->right, &newRoot);

      newRoot->left    = root->left;
      newRoot->right   = root->right;
      newRoot->balance = root->balance;

      newRoot = restoreRightBalance(newRoot, oldBalance);
    }

    if (delete) delete(root->pItem);
    free(root);
    *pStatus = OK;
    return newRoot;
  }

/**
 | Iterate on the left or the right subtree
**/

  if (c < 0) {
    if (root->left == 0) {
      *pStatus = ERROR;
      return root;
    } else {
      oldBalance = root->left->balance;
      root->left = removeNode(root->left, pTarget, compare, delete, pStatus);
      return restoreLeftBalance(root, oldBalance);
    }
  } else {
    if (root->right == 0) {
      *pStatus = ERROR;
      return root;
    } else {
      oldBalance = root->right->balance;
      root->right = removeNode(root->right, pTarget, compare, delete, pStatus);
      return restoreRightBalance(root, oldBalance);
    }
  }
}

static node *restoreLeftBalance(
  node      *root,
  const int  b
) {

  node *leftChild;

/**
 | Fix up an unbalanced left subtree, if needed
**/

  if ((leftChild = root->left) == 0) {
    ++(root->balance);
  } else if (leftChild->balance != b   &&   leftChild->balance == 0) {
    ++(root->balance);
  }
  if (root->balance > 1) return balanceNode(root);
  else                   return root;
}

static node *restoreRightBalance(
  node      *root,
  const int  b
) {
  node *rightChild;

  if ((rightChild = root->right) == 0) {
    --(root->balance);
  } else if (rightChild->balance != b   &&   rightChild->balance == 0) {
    --(root->balance);
  }
  if (root->balance < -1) return balanceNode(root);
  else                    return root;
}

static node *singleRotateLeft(
  node *root
) {
  int   balanceA;
  int   balanceB;
  node *child = root->right;

/**
 | Performs a single rotation to the left:
 | - reconnects the links;
 | - updates the balance factors.
**/

  root->right = child->left;
  child->left = root;

  balanceA = root->balance;
  balanceB = child->balance;

  if (balanceB <= 0) {
    if (balanceA >= 1) child->balance = balanceB - 1;
    else               child->balance = balanceA + balanceB - 2;
    root->balance = balanceA - 1;
  } else {
    if (balanceA <= balanceB) child->balance = balanceA - 2;
    else                      child->balance = balanceB - 1;
    root->balance = balanceA - balanceB - 1;
  }

  return child;
}

static node *singleRotateRight(
  node *root
) {
  int   balanceA;
  int   balanceB;
  node *child = root->left;

  root->left = child->right;
  child->right = root;

  balanceA = root->balance;
  balanceB = child->balance;

  if (balanceB <= 0) {
    if (balanceB > balanceA) child->balance = balanceB + 1;
    else                     child->balance = balanceA + 2;
    root->balance = balanceA - balanceB + 1;
  } else {
    if (balanceA <= -1) child->balance = balanceB + 1;
    else                child->balance = balanceA + balanceB + 2;
    root->balance = balanceA + 1;
  }

  return child;
}
