//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// AVL balanced binary tree template class.
//
// Define the preprocessor symbol C_DEBUG in order to print to the
// error stream details about every call to a ctor or a dtor.
//
// $Id: avl.hh,v 1.2 2004/04/23 06:43:51 loreti Exp $
//
// -------------------------------------------------------------------

#ifndef AVL_HH_
#define AVL_HH_

#include "bt.hh"

// Forward declarations

template <typename T> class bTree;
template <typename T> class treeSequencer;

template <typename T> class avlTree;

// A node in the balanced binary tree.  Inherits from bTree<T> so
// that, e.g., the bTree iterator works for AVL trees too; but
// redefines the internal procedures as setLeft and so on to access
// the correct data fields.  A new private data word is added, to
// store the total balance factor of the node.

template <typename T> class avlNode : public btNode<T> {
  friend class avlTree<T>;

private:
  // Data fields

  int balance_;
  avlNode * avlLeft_;
  avlNode * avlRight_;

  // Inspectors/accessors

  avlNode * getLeft() const       { return avlLeft_; }
  avlNode * getRight() const      { return avlRight_; }
  void      setLeft(avlNode * l)  { avlLeft_ = l; }
  void      setRight(avlNode * r) { avlRight_ = r; }

  // Key procedures, called from the avlTree<T> methods

  avlNode * find (const T & t) const;
  avlNode * insert(const T & t, avlNode * & pNode, bool * newNode = 0);
  avlNode * remove(const T & t, avlNode * & pNode);

  // Internal procedures used to modify and/or rebalance subtrees

  avlNode * balance();
  avlNode * singleRotateLeft();
  avlNode * singleRotateRight();
  avlNode * removeLeftmost(avlNode * & pNode);
  avlNode * restoreLeftBalance(int b);
  avlNode * restoreRightBalance(int b);

  // Debug printout

#ifdef C_DEBUG
  void printData() const {
    btNode<T>::printData();
    cerr << " b=" << balance_;
  }
#endif // C_DEBUG

  // Internal procedures that duplicate or remove a node

  void clone(const avlNode & rhs) {
    setLeft(  rhs.getLeft()  ? new avlNode( *(rhs.getLeft()) )  : 0 );
    setRight( rhs.getRight() ? new avlNode( *(rhs.getRight()) ) : 0 );
    balance_ = rhs.balance_;
    btNode<T>::clone(rhs);
  }

  void cleanup() {
    delete avlLeft_;
    delete avlRight_;
    avlLeft_ = avlRight_ = 0;
    btNode<T>::cleanup();
  }

  // Copy constructor

  avlNode(const avlNode & rhs) {
#ifdef C_DEBUG
    cerr << "avlNode copy ctor for ";
    rhs.printAll();
#endif // C_DEBUG
    clone(rhs);
  }

  // Assignment operator

  const avlNode & operator = (const avlNode & rhs) {
    if (this != &rhs) {
#ifdef C_DEBUG
      cerr << "avlNode assignment operator for ";
      rhs.printAll();
#endif // C_DEBUG
      cleanup();
      clone(rhs);
    }
    return *this;
  }

public:
  // Public constructors and destructor

  avlNode() : balance_(0), avlLeft_(0), avlRight_(0) {
#ifdef C_DEBUG
    cerr << "Default avlNode ctor\n";
#endif
  }

  avlNode(const T & t)
    : btNode<T>(t), balance_(0), avlLeft_(0), avlRight_(0) {
#ifdef C_DEBUG
    cerr << "Reference avlNode ctor for " << t << endl;
#endif // C_DEBUG
  }

  avlNode(const T * pT)
    : btNode<T>(pT), balance_(0), avlLeft_(0), avlRight_(0) {
#ifdef C_DEBUG
    cerr << "Pointer avlNode ctor for " << *pT << endl;
#endif // C_DEBUG
  }

  ~avlNode() {
#ifdef C_DEBUG
    cerr << "avlNode dtor for ";
    printAll();
#endif // C_DEBUG
    cleanup();
  }

  T & getData() const { return btNode<T>::getData(); }
};

// An AVL balanced binary tree, inheriting from bTree<T>

template <typename T> class avlTree : public bTree<T> {
private:
  // Data

  avlNode<T> * avlRoot_;

  // Inspectors/accessors

  avlNode<T> * getRoot() const             { return avlRoot_;  }
  void         setRoot(avlNode<T> * pNode) { avlRoot_ = pNode; }

  // Internal procedure to duplicate a tree

  void clone(const avlTree & rhs) {
    setRoot( rhs.getRoot() ? new avlNode<T>( *(rhs.getRoot()) ) : 0 );
  }

public:
  // Constructors and destructor

  avlTree() : avlRoot_(0) {
#ifdef C_DEBUG
    cerr << "avlTree default ctor\n";
#endif // C_DEBUG
  }

  ~avlTree() {
#ifdef C_DEBUG
    cerr << "avlTree dtor for ";
    if (getRoot()) getRoot()->printAll();
    else           cerr << "nil\n";
#endif // C_DEBUG
    delete avlRoot_;
  }

  avlTree(const avlTree & rhs) {
#ifdef C_DEBUG
    cerr << "avlTree copy ctor for ";
    if (rhs.getRoot()) rhs.getRoot()->printAll();
    else               cerr << "nil\n";
#endif // C_DEBUG
    clone(rhs);
  }

  const avlTree & operator = (const avlTree & rhs) {
    if (this != &rhs) {
#ifdef C_DEBUG
      cerr << "avlTree assignment operator for ";
      if (rhs.getRoot()) rhs.getRoot()->printAll();
      else               cerr << "nil\n";
#endif // C_DEBUG
      delete avlRoot_;
      clone(rhs);
    }
    return *this;
  }

  // Public methods to locate, insert or remove a node

  avlNode<T> * find(const T & t) const {
    if (getRoot()) return getRoot()->find(t);
    else           return 0;
  }

  avlNode<T> * insert(const T & t, bool * newNode = 0);

  bool remove(const T & t);
};

// Methods for avlNode<T>

template <typename T>
avlNode<T> * avlNode<T>::find (
  const T & t
) const {

  // Tries to find a node with a given data field in the subtree under
  // a given node.  Returns 0 if the node does not exist, or a pointer
  // to it.

  if (t < getData()) {
    if (getLeft()) return getLeft()->find(t);
  } else if (getData() < t) {
    if (getRight()) return getRight()->find(t);
  } else {
    return const_cast< avlNode<T> * >(this);
  }

  // Here if the node does not exist

  return 0;
} // find()

template <typename T>
avlNode<T> * avlNode<T>::insert(
  const T &   t,
  avlNode * & pNode,
  bool *      newNode
) {

  // Tries to insert a node with the supplied data field (t) in the
  // subtree lying under a given node, if it does not exist already.
  // Returns in "pNode" a pointer to the node holding the wanted
  // value; and (if "newNode" is not NULL) stores in *newNode 'true'
  // if a new node has been created, or 'false' if a node holding the
  // value "t" was already present.  The return value of the function
  // is a pointer to the node that has replaced the current one after
  // a rebalance.

  if (t < getData()) {

    // Insert in the left subtree

    if (getLeft()) {
      int oldBalance = getLeft()->balance_;
      setLeft( getLeft()->insert(t, pNode, newNode) );
      if (getLeft()->balance_ != oldBalance   &&   getLeft()->balance_) {
        --balance_;
      }
    } else {
      setLeft( pNode = new avlNode<T>(t) );
      if (newNode) *newNode = true;
      --balance_;
    }

  } else if (getData() < t) {

    // Insert in the right subtree

    if (getRight()) {
      int oldBalance = getRight()->balance_;
      setRight( getRight()->insert(t, pNode, newNode) );
      if (getRight()->balance_ != oldBalance   &&   getRight()->balance_) {
        ++balance_;
      }
    } else {
      setRight( pNode = new avlNode<T>(t) );
      if (newNode) *newNode = true;
      ++balance_;
    }

  } else {

    // Found an already existing node

    if (newNode) *newNode = false;
    pNode = this;
  }

  // If the tree has gone unbalanced, restore the balance

  if (balance_ < -1   ||   balance_ > 1) return balance();
  return this;
}

template <typename T>
avlNode<T> * avlNode<T>::balance() {

  // Rebalance a subtree that has gone out of balance

  if (balance_ < 0) {
    if (getLeft()->balance_ <= 0) {

      // Single rotation

      return singleRotateRight();
    } else {

      // Double rotation

      setLeft( getLeft()->singleRotateLeft() );
      return singleRotateRight();
    }

  } else {
    if (getRight()->balance_ >= 0) {

      // Single rotation

      return singleRotateLeft();
    } else {

      // Double rotation

      setRight( getRight()->singleRotateRight() );
      return singleRotateLeft();
    }
  }
}

template <typename T>
avlNode<T> * avlNode<T>::singleRotateLeft()
{
  // Performs a single rotation to the left on the current node

  avlNode<T> * nodeA = this;
  avlNode<T> * nodeB = getRight();

  // Reconnect links

  nodeA->setRight( nodeB->getLeft() );
  nodeB->setLeft( nodeA );

  // Update the balance factors

  int balanceA = nodeA->balance_;
  int balanceB = nodeB->balance_;

  if (balanceB <= 0) {
    if (balanceA >= 1) nodeB->balance_ = balanceB - 1;
    else               nodeB->balance_ = balanceA + balanceB - 2;
    nodeA->balance_ = balanceA - 1;
  } else {
    if (balanceA <= balanceB) nodeB->balance_ = balanceA - 2;
    else                      nodeB->balance_ = balanceB - 1;
    nodeA->balance_ = balanceA - balanceB - 1;
  }
  return nodeB;
}

template <typename T>
avlNode<T> * avlNode<T>::singleRotateRight()
{
  // Performs a single rotation to the right on the current node

  avlNode<T> * nodeA = this;
  avlNode<T> * nodeB = getLeft();

  nodeA->setLeft( nodeB->getRight() );
  nodeB->setRight( nodeA );

  int balanceA = nodeA->balance_;
  int balanceB = nodeB->balance_;

  if (balanceB <= 0) {
    if (balanceB > balanceA) nodeB->balance_ = balanceB + 1;
    else                     nodeB->balance_ = balanceA + 2;
    nodeA->balance_ = balanceA - balanceB + 1;
  } else {
    if (balanceA <= -1) nodeB->balance_ = balanceB + 1;
    else                nodeB->balance_ = balanceA + balanceB + 2;
    nodeA->balance_ = balanceA + 1;
  }
  return nodeB;
}

template <typename T>
avlNode<T> * avlNode<T>::remove(
  const T &      t,
  avlNode<T> * & pNode
) {

  // Removes a node holding a given value: the node is not effectively
  // removed, but disconnected from the tree (that is rebalanced); a
  // pointer to the disconnected node is returned in "pNode".

  int oldBalance;

  if (t == getData()) {

    // Gotcha!

    pNode = this;

    // Special case: if less than two children, return the only one
    // (or a NULL link).

    if (getLeft() == 0)  return getRight();
    if (getRight() == 0) return getLeft();

    // General case: get the leftmost child of the right child

    oldBalance = getRight()->balance_;
    avlNode<T> * newStart;
    setRight( getRight()->removeLeftmost(newStart) );

    // Connect the new root

    newStart->setLeft( getLeft() );
    newStart->setRight( getRight() );
    newStart->balance_ = balance_;
    return newStart->restoreRightBalance(oldBalance);

  } else if (t < getData()) {

    // Iterate on the left subtree

    if (getLeft() == 0) return this;

    oldBalance = getLeft()->balance_;
    setLeft( getLeft()->remove(t, pNode) );
    return restoreLeftBalance(oldBalance);
  } else {

    // Iterate on the right subtree

    if (getRight() == 0) return this;

    oldBalance = getRight()->balance_;
    setRight( getRight()->remove(t, pNode) );
    return restoreRightBalance(oldBalance);
  }
}

template <typename T>
avlNode<T> * avlNode<T>::removeLeftmost(
  avlNode * & pNode
) {
  // Find and remove the leftmost descendant of a given node

  avlNode<T> * leftChild = getLeft();

  if (leftChild == 0) {
    pNode = this;               // Gotcha!
    return getRight();
  }

  // Save the balance factor of the subtree; iterate; rebalance if
  // needed.

  int oldBalance = leftChild->balance_;
  setLeft( leftChild->removeLeftmost(pNode) );
  return restoreLeftBalance(oldBalance);
}

template <typename T>
avlNode<T> * avlNode<T>::restoreLeftBalance(
  int b
) {
  // Fix up an unbalanced left subtree, if needed

  avlNode<T> * leftChild = getLeft();

  if (leftChild == 0) {
    ++balance_;
  } else if (leftChild->balance_ != b   &&   leftChild->balance_ == 0) {
    ++balance_;
  }
  if (balance_ > 1) return balance();
  return this;
}

template <typename T>
avlNode<T> *avlNode<T>:: restoreRightBalance(
  int b
) {
  // Fix up an unbalanced right subtree, if needed

  avlNode<T> * rightChild = getRight();

  if (rightChild == 0) {
    --balance_;
  } else if (rightChild->balance_ != b   &&   rightChild->balance_ == 0) {
    --balance_;
  }
  if (balance_ < -1) return balance();
  return this;
}

// Methods for avlTree<T>

template <typename T>
avlNode<T> * avlTree<T>::insert(
  const T & t,
  bool *    newNode
) {
  // Tries to insert a node with a given value in an AVL tree; returns
  // a pointer to the node and, if "newNode" is not NULL, stores in
  // *newNode 'true' for an added node or 'false' for an already
  // existing node.

  avlNode<T> * pNode;

  if (getRoot()) {
    setRoot( getRoot()->insert(t, pNode, newNode) );
  } else {
    setRoot( pNode = new avlNode<T>(t) );
    if (newNode) *newNode = true;
  }
  return pNode;
}

template <typename T>
bool avlTree<T>::remove(
  const T & t
) {
  // Tries to remove a node with a given value in an AVL tree; returns
  // 'true' in case of success, or 'false' if a node with the needed
  // value does not exist in the tree.

  avlNode<T> * pNode = 0;

  if (getRoot()) setRoot( getRoot()->remove(t, pNode) );
  if (pNode) {
    pNode->setLeft(0);
    pNode->setRight(0);
    delete pNode;
    return true;
  } else {
    return false;
  }
}

#endif // AVL_HH_
