//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// A test program for the AVL tree class; or with a small sample of
// random data (if the preprocessor symbol DEBUG is defined) or with a
// simple word counting procedure.
//
// $Id: avlMain.cxx,v 1.2 2002/12/16 10:54:03 loreti Exp $
//
// -------------------------------------------------------------------

#ifdef DEBUG
#include <iostream>             // A simple test program
#include <vector>
#include "avl.hh"

using std::cout;
using std::endl;

#ifndef T_SIZE
#define T_SIZE 5
#endif

int main() {
  const size_t     treeSize(T_SIZE);
  std::vector<int> vi;
  avlTree<int>     bt0;
  size_t           i;
  bool             newNode;

  vi.reserve(treeSize);
  for (i = 0;   i < treeSize;   ++i) {
    int  num;

    do {
      num = rand();
      bt0.insert(num, &newNode);
    } while (!newNode);
    vi.push_back(num);
  }

  cout << "Tree:\t";
  for (i = 0;   i < treeSize;   ) {
    size_t last = i+4;
    if (last >= treeSize) last = treeSize-1;
    while (i < last) cout << vi[i++] << ", ";
    cout << vi[i++];
    if (i == treeSize) cout << '\n'; else cout << ",\n\t";
  }

  // Initialize bti from bt0, then remove every third element
  avlTree<int> bti(bt0);

  for (i = 0;   i < treeSize;   i += 3) {
    if (!bti.remove(vi[i])) {
      cout << "Removing " << vi[i] << " failed\n";
    }
  }

  // Reinsert every third element
  for (i = 0;   i < treeSize;   i += 3) {
    bti.insert(vi[i], &newNode);
    if (!newNode) {
      cout << "Reinsertion of " << vi[i] << " failed\n";
    }
  }

  treeSequencer<int> inSeq(bti);
  treeSequencer<int> rInSeq(bti,   treeSequencer<int>::INORDER,    true);
  treeSequencer<int> preSeq(bti,   treeSequencer<int>::PREORDER);
  treeSequencer<int> rPreSeq(bti,  treeSequencer<int>::PREORDER,   true);
  treeSequencer<int> postSeq(bti,  treeSequencer<int>::POSTORDER);
  treeSequencer<int> rPostSeq(bti, treeSequencer<int>::POSTORDER,  true);

  cout << "InOrder sequence, ascending and descending:\n";
  for (i = 1;   !inSeq;   ++inSeq,  ++rInSeq,  ++i) {
    cout << "\tElement " << i << ":\t" << inSeq()
         << '\t' << rInSeq() << endl;
  }

  cout << "PreOrder sequence, ascending and descending:\n";
  for (i = 1;   !preSeq;   ++preSeq,  ++rPreSeq,  ++i) {
    cout << "\tElement " << i << ":\t" << preSeq()
         << '\t' << rPreSeq() << endl;
  }

  cout << "PostOrder sequence, ascending and descending:\n";
  for (i = 1;   !postSeq;   ++postSeq,  ++rPostSeq,  ++i) {
    cout << "\tElement " << i << ":\t" << postSeq()
         << '\t' << rPostSeq() << endl;
  }

  return 0;
}
#else  // DEBUG
#include <fstream>              // A word-counting program
#include <iostream>
#include <string>
#include <cctype>
#include "avl.hh"

using std::cerr;
using std::cout;
using std::ifstream;
using std::string;

// A structure to hold a word, and the number of its occurrences

struct freq {
  string        word;
  unsigned long nTimes;
  freq(const char * pc) : word(pc), nTimes(1) {}
  bool operator <  (const freq & rhs) const { return word < rhs.word;  }
  bool operator == (const freq & rhs) const { return word == rhs.word; }
};


const char * getWord(
  ifstream & ifs
) {
  // Reads from the input stream a "word", defined as an alphabetic
  // character followed by other alphabetic characters and/or numbers
  // and/or underscores); and returns a pointer to the word itself,
  // converted to lowercase.

  static const size_t bufferLength = 128;
  static char         buffer[bufferLength];
  static size_t       next = 0;
  static bool         eof_reached = false;

  if (eof_reached) return 0;

  for (;;) {
    ifstream::int_type c;

    if ((c = ifs.get()) == ifstream::traits_type::eof()) {
      if (next) {
        eof_reached  = true;
        buffer[next] = '\0';
        return buffer;
      } else {
        return 0;
      }
    }

    c = tolower(static_cast<unsigned char>(c));
    if (!next) {
      if (isalpha(static_cast<unsigned char>(c))) buffer[next++] = c;
    } else {
      if (isalpha(static_cast<unsigned char>(c))   ||
          isdigit(static_cast<unsigned char>(c))   ||   c == '_') {
        buffer[next++] = c;
      } else {
        next = buffer[next] = '\0';
        return buffer;
      }
    }
  }
}

int main(
  int   argc,
  char *argv[]
) {
  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " <1 or more file names>\n";
  } else {
    unsigned long nWords = 0, nDiffs = 0;
    avlTree<freq> btf;

    while (--argc) {
      ifstream ifs(*++argv);

      if (ifs) {
        cout << "Working on " << *argv << " ...";
        cout.flush();

        while (const char *pc = getWord(ifs)) {
          freq          f(pc);
          bool          newNode;
          avlNode<freq> *pNode = btf.insert(f, &newNode);
          if (newNode) {
            ++nDiffs;
          } else {
            ++(pNode->getData().nTimes);
          }
          ++nWords;
        }
        ifs.close();
        cout << " done." << std::endl;
      } else {
        cerr << "Error opening file \"" << *argv << "\"\n";
      }
    }

    cerr << "\n  Found " << nWords <<" words ("
         << nDiffs << " different)\n\n";
    treeSequencer<freq> inSeq(btf);
    cout << "* In alphabetical order:\n";
    while (!inSeq) {
      cout << "- " << inSeq().word << ": " << inSeq().nTimes << " times\n";
      ++inSeq;
    }
  }
}

#endif // DEBUG
