/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 General binary tree routines: insertion, remotion, search,
 tree traversal.

 $Id: tree.c,v 1.1 2001/01/17 11:20:10 loreti Exp $

 *------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "tree.h"

/**
 | Local procedure prototypes
**/

static node *removeNode(node *, void(item *));

int insertTree(
  node **pRoot,
  item  *pItem,
  node **ppNode,
  int    compare(const item *, const item *)
) {

/**
 | "insertTree" tries to insert in the binary tree rooted at **pRoot a
 | node containing the user data in *pItem.  The return value is OK in
 | case of success; if a node with the given item already exists, the
 | tree is unchanged and the return value is ERROR.  In both cases, in
 | *pNode will be stored a pointer to the node containing the given
 | item.
 |
 | "compare" is a function taking two pointers to data items, that
 | returns a negative value if the first one is lesser that the second
 | one; a positive value if the first one is bigger than the second
 | one; or zero if the two items have the same value.
**/

/**
 | Special case: the tree is empty.
**/

  if (*pRoot == 0) {
    node *pNode;

    if ((pNode = malloc(sizeof(node))) == 0) {
      fputs("insertTree: couldn't get heap memory\n", stderr);
      exit(EXIT_FAILURE);
    }

    pNode->left  = 0;
    pNode->right = 0;
    pNode->pItem = pItem;
    *pRoot       = pNode;
    *ppNode      = pNode;
    return OK;

  } else {
    int c;

    c = compare(pItem, (*pRoot)->pItem);
    if (c < 0) {
      return insertTree(&((*pRoot)->left), pItem, ppNode, compare);
    } else if (c > 0) {
      return insertTree(&((*pRoot)->right), pItem, ppNode, compare);
    } else {
      *ppNode = *pRoot;
      return ERROR;
    }
  }
}

void removeTree(
  node *root,
  void delete(item *)
) {

/**
 | "removeTree" removes the subtree rooted at *root, calling for every
 | user data item the procedure "delete" (if supplied)
**/

  if (root) {
    removeTree(root->left, delete);
    removeTree(root->right, delete);
    if (delete) delete(root->pItem);
    free(root);
  }
}

void traverseTree(
  const node  *root,
  const order  o,
  void         fun(const item *)
) {

/**
 | "traverseTree" traverses the subtree rooted at *root, calling for
 | every user data item the supplied procedure "fun".
**/

  if (root) {
    if (o == PREORDER) fun(root->pItem);
    traverseTree(root->left, o, fun);
    if (o == INORDER) fun(root->pItem);
    traverseTree(root->right, o, fun);
    if (o == POSTORDER) fun(root->pItem);
  }
}

item *findTree(
  node       *root,
  const item *pTarget,
  int         compare(const item *, const item *)
) {

/**
 | "findTree" explores the subtree rooted at *root, searching for a
 | node comparing equal to the one in *pTarget.  Returns a pointer to
 | the node, or zero if it has not been found.
**/

  if (root) {
    int c = compare(pTarget, root->pItem);

    if (c < 0) {
      return findTree(root->left, pTarget, compare);
    } else if (c > 0) {
      return findTree(root->right, pTarget, compare);
    } else {
      return root->pItem;
    }
  } else {
    return 0;
  }
}

int removeFromTree(
  node       **pRoot,
  const item  *pTarget,
  int          compare(const item *, const item *),
  void         delete(item *)
) {

/**
 | "removeFromTree" explores the binary tree rooted at **pRoot,
 | searching for a node containing an user data item comparing equal
 | to *pTarget; the node is then removed from the tree, after having
 | called the procedure "delete" (if supplied).
**/

  node *pNode = *pRoot;
  int   c;

/**
 | Two special cases: an empty tree, and a tree with the wanted data
 | in the root node.
**/

  if (pNode == 0) return ERROR;

  if ((c = compare(pTarget, pNode->pItem)) == 0) {
    *pRoot = removeNode(pNode, delete);
    return OK;
  }

/**
 | In the general case, we slide down the tree looking for *pTarget,
 | saving the address of the parent node; the child is then removed by
 | the internal procedure "removeNode", and the parent link updated.
**/

  while (pNode) {
    node *nextNode = 0;
    int   nextC    = 0;

    if (c < 0) {
      nextNode = pNode->left;
      if (nextNode   &&   (nextC = compare(pTarget, nextNode->pItem)) == 0) {
        pNode->left = removeNode(nextNode, delete);
        return OK;
      }
    } else if (c > 0) {
      nextNode = pNode->right;
      if (nextNode   &&   (nextC = compare(pTarget, nextNode->pItem)) == 0) {
        pNode->right = removeNode(nextNode, delete);
        return OK;
      }
    }

    pNode = nextNode;
    c     = nextC;
  }
  return ERROR;
}

static node *removeNode(
  node *pNode,
  void  delete(item *)
) {

/**
 | "removeNode" removes a node from the subtree rooted at *pNode,
 | returning the address of the new root.
**/

  node *parent;
  node *retVal;

/**
 | Special cases: a tree with less that two children
**/

  if (pNode->left == 0) {
    retVal = pNode->right;
  } else {
    if (pNode->right == 0) {
      retVal = pNode->left;
    } else {
      retVal = pNode->right->left;

/**
 | Another special case: the right child has no left child
**/

      if (retVal == 0) {
        pNode->right->left = pNode->left;
        retVal = pNode->right;
      } else {

/**
 | General case: we slide down the left subtree of the right child to
 | find the leftmost node; that node is disconnected and relinked at
 | the top.
**/

        parent = pNode->right;
        while (retVal->left) {
          parent = retVal;
          retVal = retVal->left;
        }
        parent->left  = retVal->right;
        retVal->left  = pNode->left;
        retVal->right = pNode->right;
      }
    }
  }

/**
 | Now is the time to effectively remove the old top node
**/

  if (delete) delete(pNode->pItem);
  free(pNode);
  return retVal;
}
